/* A verilog implementation of the Baillie-PSW primality test.
The Baillie-PSW test consists of the following segments:
1) Check for divisiblity by some small primes.
2) Perform a Miller-Rabin test to the base 2.
3) In the sequence: 5,-7,9,-11,...  , find D such that (D/N) = -1 , where (a/b) is the Jacobi symbol
4) Using D from step 3, perform a standard/strong Lucas primality test using Selfridge parameters(P=1,Q=(1-D)/4).
If a number passes all the tests(1,2 and 4), then it can be strongly concluded that the number is prime.

The strength of the Baillie-PSW test arises from the fact that there is little/no overlap between the "base 2 Miller-Rabin psuedoprimes" and the "Lucas-Selfridge psuedoprimes".
Due to this fact, for any number < (2^64), Baillie-PSW does not have any psuedoprimes. This value remains small for numbers > (2^64) as well.

*/

module bpsw #(
	parameter SLICE_LEN = 200,			//Size of input/output port
	parameter NLEN = 1000				//Size of number to be tested
) (
	input [SLICE_LEN-1:0] N_slice,		//250 bit input port to receive initial random number
	input in_ready,						//Input signal to indicate completion of RN-slice transmission
	input clk,							//Global clock signal
	input reset,						//Input signal to reset module
	output reg in_ack=0,				//Output signal to indicate completion of RN-slice reception
	output out,							//Output signal to indicate primality : 0 - composite, 1 - prime
	output reg out_ready=0,				//Output signal to indicate ready status of output
	output reg error=0,					//Output signal to indicate any error encountered during computation
	output reg [SLICE_LEN-1:0] N_out	//Port to output final prime number
);

localparam INITVAL = 5'b11111;
reg signed [NLEN:0] N = 0;				//number to be tested(formed by concatenating all N_slice signals)
//wire [2:0] outs,out_readys;				//Indicate values of out and out_ready signals of each sub-module
reg signed [4:0] state=INITVAL;			//State variable
wire outs_ready;						//Indicates status of output of each sub-module: 1 - all outputs ready
integer i = 0;							//Iteration variable


//////////////////////////////SUB-MODULE INSTANTIATIONS//////////////////////////////

/*******************************************
Signals connected to sub-modules:
div_	concerns divisibility module signals
mr_ 	concerns miller-rabin module signals
luc_ 	concerns lucas module signals
*******************************************/
reg div_in_ready=0,mr_in_ready=0,luc_in_ready=0;	//Input signals indicating module to start computation
reg div_reset=0,mr_reset=0,luc_reset=0;				//Input signals indicating module to reset
reg mr_abort=0,luc_abort=0;							//Input signals indicating module to abort computation
wire div_out_ready,mr_out_ready,luc_out_ready;		//Output signals indicating ready status of output
wire div_out,mr_out,luc_out;						//Output signals indicating output of module


//Divisibility module
divisibility #(
.NLEN 		( NLEN			)
) div (
.N 			( N				),
.in_ready 	( div_in_ready	),
.clk 		( clk			),
.reset 		( div_reset		),
.out 		( div_out		),
.out_ready 	( div_out_ready	)
);

//Miller-Rabin module instantiation
mr #(
.NLEN 		( NLEN			)
) millrab (
.N 			( N				),
.in_ready 	( mr_in_ready	),
.clk 		( clk			),
.reset 		( mr_reset		),
.out 		( mr_out		),
.out_ready 	( mr_out_ready	),
.abort 		( mr_abort		)
);

//Lucas module instantiation
lucas #(
.NLEN 		( NLEN			)
) luc (
.N 			( N				),
.in_ready 	( luc_in_ready	),
.clk 		( clk			),
.reset 		( luc_reset		),
.out 		( luc_out		),
.out_ready 	( luc_out_ready	),
.abort 		( luc_abort		)
);

//////////////////////////SUB-MODULE INSTANTIATIONS COMPLETE//////////////////////////


always @(posedge clk)
begin
	
	if(reset==1) begin
		state <= INITVAL;
		out_ready <= 0;
	end
	else begin
		
		case(state)
		
			INITVAL:	begin
							error <= 0;
							i <= 0;
							out_ready <= 0;
							if(in_ready==1)
								state <= 0;
							N[NLEN] <= 1'b0;
							in_ack <= 0;
						end
			
			0 :	begin
					state <= 1;
					//logic to concatenate slices to form complete random number
					if(i==0) begin
						N[SLICE_LEN-1:0] <= N_slice[SLICE_LEN-1:0];
					end
					else if(i==1) begin
						N[2*SLICE_LEN-1:SLICE_LEN] <= N_slice[SLICE_LEN-1:0];
					end
					else if(i==2) begin
						N[3*SLICE_LEN-1:2*SLICE_LEN] <= N_slice[SLICE_LEN-1:0];
					end
					else if(i==3) begin
						N[4*SLICE_LEN-1:3*SLICE_LEN] <= N_slice[SLICE_LEN-1:0];
					end
					else if(i==4) begin
						N[5*SLICE_LEN-1:4*SLICE_LEN] <= N_slice[SLICE_LEN-1:0];
					end
					else begin
						$display("Error in random number reception");			//FOR TESTING PURPOSES ONLY/////////////////////////////////
						error <= 1;
					end
				end
			
			1 : begin
					if(i<4) begin
						i <= i + 1;
						in_ack <= 1;
						state <= 2;
					end
					else begin
						in_ack <= 1;
						state <= 3;
						$display("Initial Random Number = %d",N);				//FOR TESTING PURPOSES ONLY/////////////////////////////////
						if(N[0]==1'b0)				//If number is even, make it odd
							N[0] <= 1'b1;
					end
				end
			
			2 : begin
					in_ack <= 0;
					if(in_ready==1) begin
						state <= 0;
					end
				end
				
			3 : begin
					if(in_ready==1) begin
						$display("Yes");										//FOR TESTING PURPOSES ONLY/////////////////////////////////
					div_in_ready <= 1;
					mr_in_ready <= 1;
					luc_in_ready <= 1;
					state <= 4;
					end
				end
				
			4 : begin
					//If divisibility module declares number composite, then abort miller-rabin and lucas modules
					if(div_out_ready==1) begin
						if(div_out == 0) begin		
							mr_abort <= (mr_out_ready==1) ? 1'b0 : 1'b1;
							luc_abort <= (luc_out_ready==1) ? 1'b0 : 1'b1;
						end
						state <= 5;
					end
				end
				
			5 : begin
					if(outs_ready==1) begin
						if(out==1) begin			//number is prime - start prime transmission sequence
							i <= 0;
							state <= 7;
							$display("B : %d is prime",N);						//FOR TESTING PURPOSES ONLY/////////////////////////////////
						end
						else begin					//number is composite - add 2 and continue search
							mr_abort <= 1'b0;
							luc_abort <= 1'b0;
							div_reset <= 1;
							mr_reset <= 1;
							luc_reset <= 1;
							N <= N + 2;
							state <= 6;
						end
					end
				end
			
			6 : begin
					if(N[NLEN]==1'b1)				//If number overflows, then bring number back to range
						N[NLEN] <= 1'b0;
					div_reset <= 0;
					mr_reset <= 0;
					luc_reset <= 0;
					state <= 3;
				end
				
			7 : begin
					out_ready <= 1;
					state <= 8;
					//logic to transmit found prime
					if(i==0) begin
						N_out <= N[SLICE_LEN-1:0];
					end
					else if(i==1) begin
						N_out <= N[2*SLICE_LEN-1:SLICE_LEN];
					end
					else if(i==2) begin
						N_out <= N[3*SLICE_LEN-1:2*SLICE_LEN];
					end
					else if(i==3) begin	
						N_out <= N[4*SLICE_LEN-1:3*SLICE_LEN];
					end
					else if(i==4) begin
						N_out <= N[5*SLICE_LEN-1:4*SLICE_LEN];
					end
					else begin
						$display("Error in prime number transmission");			//FOR TESTING PURPOSES ONLY/////////////////////////////////
						error <= 1;
					end
				end
				
			8 : begin
					if(i<4) begin
						i <= i + 1;
						state <= 7;
					end
				end
				
			default : 	begin
							$display("Invalid state entered");					//FOR TESTING PURPOSES ONLY/////////////////////////////////
							error <= 1;
						end
				
		endcase
		
	end
	
end	


//Only if all 3 modules declare N to be prime(output as 1), the main module declares the number to be prime
assign out = div_out && mr_out && luc_out;

//When the outputs of all 3 modules are ready, only then the main module declares output to be ready
assign outs_ready = div_out_ready && mr_out_ready && luc_out_ready;

endmodule